% Graph G 1 is isomorph to a subgraph of G 2 for directed graphs - with labeled edges
% and with labeled nodes
%
%

% subgraphisomorphy( + graph1, +  graph2, - List with nodes of graph 1 and their matched nodes in graph 2 )
% Graphsyntax: 
%       graph( <GraphName>, [ rel( <Relationtype>, node( <Nodetype>, < Nodename>), 
 %                                                                             node( <Nodetype>, < Nodename>) ) * ] )

subgraphisomorphy( graph( G1, RelList1), graph( G2, RelList2), ListOfNodes1New) :-
	graphTotalInfo( G1, NumberOfNodes1, ListOfNodes1, RelList1),
	graphTotalInfo( G2, NumberOfNodes2, ListOfNodes2, RelList2),
             changeNodeNames( ListOfNodes1, ListOfNodes1New, RelList1, RelList1New), !,
	isomorph( graph( G1, NumberOfNodes1, ListOfNodes1New, RelList1New),
		    graph( G2, NumberOfNodes2, ListOfNodes2, RelList2 ) ).


% Checks, whether the first graph is subgraphisomorph / graphisomorph with the second graph 
% The result is written into the first graph 
% isomorph( graph(  + NameOfGraph 1, + Number of nodes of Graph 1, + List of nodes, 
%                               + List of Graphrelations),
%                  graph(  + NameOfGraph 2, + Number of nodes of Graph 2, + List of nodes, 
%                               + List of Graphrelations) )

isomorph( graph( _Name1 , NumberOfNodes1, ListOfNodes1New, RelList1),
                 graph( _Name2, NumberOfNodes2, ListOfNodes2,         RelList2) ) :-
	    specialTests( NumberOfNodes1, NumberOfNodes2 ) ,
	    isomorph( ListOfNodes1New, RelList1, ListOfNodes2, RelList2). 

% specialTests( + Number, + Number )
specialTests( NumberOfNodes1, NumberOfNodes2 ):-
	NumberOfNodes1 =<  NumberOfNodes2 .

%isomorph( + list of nodes of graph 1, +  list of relations of graph 1, 
%                + list of nodes of graph 2, + list of relations of graph 2). 
isomorph( [ ], RelList1, _ListOfNodes, RelList2) :-
	relationsAreSubset( RelList1, RelList2). 

isomorph( [ Node | ListOfNodes1 ], RelationList1, ListOfNodes2, RelList2) :- 
	select( Node, ListOfNodes2, _NodeSelected, RestOfNodesList2),
	isomorph( ListOfNodes1, RelationList1, RestOfNodesList2, RelList2).


% A equivalence in the graph G2 is selected for the node NodeOld in the graph G1.
% The new selected equivalent node replaces all old entries for the node NodeOld in graph G1
% select( node( + number of relations of the node, + type of node, 
%                       ( + node nameof graph 1, - node name of graph 2) ),
%                      + list of nodes of graph 2)

select( node( Num1, Type, ( _NodeOld, NodeNew) ), [ node( Num2, Type, NodeNew) | RestOfNodesList], 
           node( Num2, Type, NodeNew), RestOfNodesList ) :-
                    Num1 =< Num2.

select( Node1,                               [ Node2 | ListNew ], NodeSelected, [ Node2  | RestOfNodesList ] ) :-
	select( Node1, ListNew, NodeSelected, RestOfNodesList ).


% relationsAreSubset( + list of relation of graph 1 filled mit matching nodes of graph2, 
%                                 + list of relation of graph 2)
relationsAreSubset( [ ] , _RelList) .
relationsAreSubset( [ rel( Label, node( Type1, (_NodeOld1, NodeNew1) ),
                                                    node( Type2, (_NodeOld2, NodeNew2) ) )|       
                                   RelListGraphOne],  RelListGraphTwo ) :-
	memberDelete( rel(Label, node( Type1, NodeNew1), node( Type2, NodeNew2)) , 
                                        RelListGraphTwo, RelListGraphTwoNew ),
	relationsAreSubset( RelListGraphOne, RelListGraphTwoNew).



% checks whether a relation with a special label from node 1 to node 2 is member of the relation list
% if it is a member it deletes and the predicate is true
% memberDelete( + rel(Label, Node1, Node2) , + list of relations rel( Label, 
%                         - list of relation without the asked relation )
memberDelete( rel(Label, Node1, Node2) , [ rel( Label, Node1, Node2) |   RelList ], RelList ) :- !.
memberDelete( rel(Label, Node1, Node2), [ Relation | RelList ], [ Relation | RelListNew ] ) :-
                memberDelete( rel(Label, Node1, Node2), RelList, RelListNew ).
	


% Result: Number of the nodes in the graph, the node are sorted by their number of relations 
% graphTotalInfo( + string, - number, - list, + graph by relations).

graphTotalInfo( _Name, NumberOfNodes, ListOfNodesWithRelationNumbers, RelationList) :-
	countAndSeperateNodes( 0, NumberOfNodes, [ ], ListOfNodes, RelationList),
	countRelationsperNode( ListOfNodes, RelationList , ListOfNodesWithRelationNumbers ) .
	
% seperates the nodes of the graph and count der nodes
% countAndSeperateNodes( + number, - number, + list, - list, + list ).

countAndSeperateNodes( NumberResult, NumberResult, ListOfNodesResult, ListOfNodesResult, [   ] ).
countAndSeperateNodes( NumberOld, NumberResult, 
                                         ListOfNodes, ListOfNodesResult, 
                                         [ rel( _Label, node( Type1, Node1),  node( Type2,  Node2))  | RestOfRelations ] ) :-
	memberNodes( [ node(Type1, Node1), node( Type2, Node2) ] , Number, ListOfNodes,       
                                         ListOfNodesNew),
	append(ListOfNodes, ListOfNodesNew, ListOfNodesAll),
	NumberNew is NumberOld + Number,
	countAndSeperateNodes( NumberNew, NumberResult, 
                                                      ListOfNodesAll, ListOfNodesResult, RestOfRelations ).

% extract the number of relations per node
% countRelationsperNode( + list, + list, - list) 
countRelationsperNode( [ ] , _RelationList, [ ] ).
countRelationsperNode( [ node( Type, NodeName) | NodeRelationList ], RelationList, 
                                        [ node( Number, Type, NodeName) | ListOfNodes ] ) :-
	countRelations( NodeName, Number, RelationList ) ,
	countRelationsperNode( NodeRelationList , RelationList, ListOfNodes ).

countRelations( _ , 0, [ ] ).
countRelations( NodeName, NumberNew, [ Relation | RelationList ] ) :-
	check( NodeName, Relation, NumberFirst) ,
	countRelations( NodeName, NumberSecond, RelationList ),
	NumberNew is NumberFirst + NumberSecond .

check( NodeName,   rel( _Label, node( _Type, NodeName), _ ), 1 ) :- !.
check( NodeName,   rel( _Label, _ , node( _Type, NodeName)    ), 1 ) :- !.
check( _NodeName, rel( _Label, _ , _ ), 	  0 ).



% Check whether node are new and which of  them are new 

memberNodes( [ ], 0 , _ListOfNodes, [ ] ).
memberNodes( [ Node | Rest ] , NumberNew,  ListOfNodes, ResultListOfNewNodes ) :-
	memberNodes( Rest, NumberOld, ListOfNodes, ListOfNodesNew),
 	memberNode( Node, ListOfNodes, ListWithNewNode, Number) ,
	append( ListWithNewNode, ListOfNodesNew, ResultListOfNewNodes), 
NumberNew is NumberOld + Number.
	
	
memberNode( Node, List, [ ] , 0 ):-
	member( Node, List ), !.

memberNode( Node, _List, [ Node ] , 1 ).

% changeNodeNames( + list, - list, + list, - list)
changeNodeNames( [ ] , [ ], Result, Result  ).
changeNodeNames( [ node( Num, Type, NodeName) | NodesRest ] , 
	                     [ node( Num, Type, (NodeName, _NodeNameNew) )| NodesRestNew ] , 
                                   RelationList , RelationListResult ) :-
	 changeRelationList( NodeName, _NodeNameNew, RelationList, RelationListNew ),
	 changeNodeNames( NodesRest, NodesRestNew,  RelationListNew, RelationListResult ). 

% changeRelationList( + string, + variable, RelationList, RelationListNew )
changeRelationList( _NodeName, _NodeNameNew, [ ] , [ ] ).
changeRelationList( NodeName, NodeNameNew, 
                                [ rel( Label, node( Type, NodeName), NodeNameOther )| RelationList ],
                                [ rel( Label, node(Type, (NodeName, NodeNameNew)), NodeNameOther )| 
                                  RelationListNew ] )  :- !,
              changeRelationList( NodeName, NodeNameNew, RelationList , RelationListNew).
	
changeRelationList( NodeName, NodeNameNew, 
                                 [ rel( Label, NodeNameOther, node( Type, NodeName) )| RelationList ],
                                 [ rel( Label, NodeNameOther, 
                                   node( Type, (NodeName, NodeNameNew)) ) | RelationListNew ] )  :- !,
              changeRelationList( NodeName, NodeNameNew, RelationList , RelationListNew).

changeRelationList( NodeName, NodeNameNew, 
		     [ rel( Label, NodeNameOther1, NodeNameOther2 ) | RelationList ],
                                 [ rel( Label, NodeNameOther1, NodeNameOther2 ) | RelationListNew ] )  :- !,
              changeRelationList( NodeName, NodeNameNew, RelationList , RelationListNew).


